/***************************************************************************
 *   Copyright (C) 2005 by Siegfried Nijssen                               *
 *   snijssen@informatik.uni-freiburg.de                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef GGRAPH_H
#define GGRAPH_H
#include <vector>
#include <istream>
#include <ostream>
#include "global.h"
#include "ggraphcode.h"

struct GGraphEdge {
    GEdgeData data;
    GNodeID nodeid;
};

struct GGraphNode {
    GNodeData data;
    vector<GGraphEdge> edges;
};

struct GGraphExtraData {
  // by default no extra data. This class can be derived.
};

/**
@author Siegfried Nijssen
*/
class GGraph{
public:
    GGraph();

    ~GGraph();
    void addNode ( GNodeData nodedata );
    void delNode ();
    void addEdge ( GNodeID from, GNodeID to, GEdgeData edgedata );
    void delEdge ( GNodeID from, GNodeID to );
    bool isEdge ( GNodeID from, GNodeID to, GEdgeData data );
    bool isSubgraphOf ( GGraph *graph );
        
    vector<GGraphNode> nodes;
    
    GGraphCode *code; // a code for this graph; NULL if not available. Code is not automatically updated.
    GGraphExtraData *extradata;
};

istream& operator>>(istream& stream, GGraph &graph );
ostream& operator<<(ostream& stream, GGraph &graph );

#endif
