/***************************************************************************
 *   Copyright (C) 2005 by Siegfried Nijssen                               *
 *   snijssen@informatik.uni-freiburg.de                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "glattice.h"

GLattice::GLattice()
{
}


GLattice::~GLattice()
{
}

void GLattice::print()
{
  /**
   * Will start the recursive calls of print() to print the lattice/trie
   */
  print(0, &roots);
}

void GLattice::addNode(vector<GLatticeNode*> &latticetrace, GSearchCode *searchcode, vector<GFrequency> &frequency )
{
  /**
   * Add a node (ie Graph) to the lattice and construct the according links in the lattice
   */
  pair<GLatticeNode*,int> p2, p = insert (	searchcode->tokens, 
						latticetrace[searchcode->tokens.size ()-1],
						searchcode->tokens.size () - 2 );
  latticetrace.push_back ( p.first );
  p.first->value.frequency = frequency;
  static int cnt = 0;
  p.first->value.index = cnt;
  cnt++;

  for ( int i = 0, size = searchcode->tokens.size () - 1; i < size; ++i ) 
    if ( searchcode->cycletokens[i] ||  // is this edge in a cycle?
         ( i == 0 && searchcode->graph.nodes[0].edges.size () == 1 ) || // is this edge a leaf?
         ( searchcode->tokens[i].direction == forward &&  // is this edge a leaf?
           searchcode->tokens[i].to != searchcode->tokens[i+1].from ) ) {
      searchcode->graph.delEdge ( searchcode->tokens[i].from, searchcode->tokens[i].to );
      GGraphCode *code = GCanonicalDFSComputer ( &searchcode->graph ).computeCanonical ();

      p2 = insert ( code->tokens, latticetrace[i], i - 1 );	// This should fail

      addEdge ( p.first->value.subgraphs, p2.first );
      addEdge ( p2.first->value.supergraphs, p.first );

      delete code;

      searchcode->graph.addEdge ( searchcode->tokens[i].from, searchcode->tokens[i].to, searchcode->tokens[i].data );
    }

    if ( searchcode->tokens.size () > 1 ) {
      addEdge ( latticetrace.back ()->value.subgraphs, latticetrace[latticetrace.size () - 2] );
      addEdge ( latticetrace[latticetrace.size () - 2]->value.supergraphs, latticetrace.back () );
    }
}

// Private functions

void GLattice::print(int v, vector<GLatticeNode *> *nodes)
{
  /**
  * recursive print() function to print the lattice/trie
  */
  for ( vector<GLatticeNode *>::iterator p = nodes->begin (); p != nodes->end (); ++p ) {
    printf ( "%i ", (*p)->value.index );
    for ( int i = 0; i < v; ++i )
      printf ( " " );
    printGToken ( (*p)->key );
    printf ( " (%i,%i)", (*p)->value.frequency[0], (*p)->value.frequency[1] );
    for ( int i = 0, size = (*p)->value.supergraphs.size (); i < size; ++i )
      printf ( " %i", (*p)->value.supergraphs[i]->value.index );
    printf ( "\n" );
    print ( v + 2, &(*p)->children ); // recursion
  }

}

void GLattice::addEdge ( vector<GLatticeNode*> &nodes, GLatticeNode *node )
{
  /**
  * will add a edge from every node in nodes to node
  */
  vector<GLatticeNode*>::iterator v;
  v = lower_bound ( nodes.begin (), nodes.end (), node );
  if ( v == nodes.end () || *v != node )
    nodes.insert ( v, node );
}
















