/***************************************************************************
 *   Copyright (C) 2005 by Siegfried Nijssen                               *
 *   snijssen@informatik.uni-freiburg.de                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "gresultset.h"
#include <iostream>

GResultSet::GResultSet()
{
  minchi2 = 0.0;
  size = 0;
  shrinkcount = 0;
}


GResultSet::~GResultSet()
{
}


void GResultSet::insert ( double chi2, vector<GFrequency> &frequency, GGraph *graph ) {
  // this function determines if and where in the result set the pattern is stored.
  if ( !results.empty () ) {
    if ( minchi2 > chi2 )
      return;
    pair<multimap<double,GResult>::iterator,multimap<double,GResult>::iterator> p = results.equal_range ( chi2 );
    if ( p.first != results.end () ) {
      bool first = false;
      for ( ; p.first != p.second; ) {
        bool second = false;
        if ( p.first->second.graph.nodes.size () < graph->nodes.size () ) {
          if ( p.first->second.graph.isSubgraphOf ( graph ) ) {
            return; // the new graph is just as good as the old one, prefer the old one as it's smaller
          }
        }
        else
          if ( p.first->second.graph.nodes.size () > graph->nodes.size () )  {
            if ( graph->isSubgraphOf ( &p.first->second.graph ) ) 
              if ( first ) {
                GGraphCode *c = p.first->second.graph.code;
                p.first->second.graph = *graph;
                p.first->second.graph.code = c;
                p.first->second.graph.code->tokens = graph->code->tokens; // replace the old graph with the new, smaller one
                first = true;
              }
              else {
                multimap<double,GResult>::iterator del = p.first;
                ++p.first;
                results.erase ( del );
                ++shrinkcount;
                --size; // we're shrinking
                second = true;
              }
          }
        if ( !second )
          ++p.first;
      }
      if ( first )
        return;
    }
    if ( size == maxsize ) 
      results.erase ( results.begin () );
    else 
      if ( minchi2 == chi2 ) // no preference for equal chi2s that are not subgraphs of each other
        return;
      else
        ++size; // we're going to grow
  }
  else
    ++size; // we're going to grow
  multimap<double,GResult>::iterator it2 = results.insert ( make_pair ( chi2, GResult () ) );
  it2->second.graph = *graph;
  it2->second.graph.code = new GGraphCode;
  it2->second.graph.code->tokens = graph->code->tokens;
  it2->second.chi2 = chi2;
  it2->second.frequency = frequency;
  minchi2 = results.begin ()->first;
}

ostream &operator<< ( ostream &stream,GResultSet &resultset ) {
  for ( multimap<double,GResult>::iterator it = resultset.results.begin ();
                                    it != resultset.results.end ();
                                    ++it )
    stream << "# " << it->second.frequency[0] << " " << it->second.frequency[1] << " " << it->first << endl 
           << it->second.graph << endl;
  
  return stream;
}