/***************************************************************************
 *   Copyright (C) 2005 by Siegfried Nijssen                               *
 *   snijssen@informatik.uni-freiburg.de                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef GRESULTSET_H
#define GRESULTSET_H
#include <map>
#include "gsearchcode.h"

struct GResult {
  GGraph graph;
  double chi2; // parts of the result, add as desired
  vector<GFrequency> frequency;
};

/**
@author Siegfried Nijssen
*/
class GResultSet{
    friend ostream &operator<< ( ostream &stream, GResultSet &resultset );
private:
    multimap<double,GResult> results;
    double minchi2; // the smallest chi2 stored in the set
    int maxsize, // the maximum size of the result set
        size; // the size of the current set.
public:    
    int shrinkcount;
    GResultSet();

    ~GResultSet();
    
    void setSize ( int size ) { this->maxsize = size; }
    double minChi2 () const { return minchi2; }
    void setMinChi2 ( double val ) { minchi2 = val; } 
    void insert ( double chi2, vector<GFrequency> &frequency, GGraph *graph ); 
};

ostream &operator<< ( ostream &stream, GResultSet &resultset );

#endif
