#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import networkx as nx
import numpy as np
import collections
from tkinter import *
#from sklearn import metrics

import matplotlib
matplotlib.use('TkAgg')
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg

import matplotlib.pyplot as plt

from sklearn import tree
import openbabel, pybel
#from sklearn.ensemble import RandomForestClassifier
#from sklearn import svm
#from sklearn.naive_bayes import GaussianNB
#from sklearn import linear_model
#from sklearn.linear_model import SGDClassifier

CHEM_EL = {1:{'name':'H','color':'yellow'},
2:{'name':'He','color':'grey'},
3:{'name':'Li','color':'grey'},
4:{'name':'Be','color':'grey'},
5:{'name':'B','color':'grey'},
6:{'name':'C','color':'orange'},
7:{'name':'N','color':'red'},
8:{'name':'O','color':'green'},
9:{'name':'F','color':'blue'},
10:{'name':'Ne','color':'grey'},
11:{'name':'Na','color':'grey'},
12:{'name':'Mg','color':'grey'},
13:{'name':'Al','color':'grey'},
14:{'name':'Si','color':'grey'},
15:{'name':'P','color':'grey'},
16:{'name':'S','color':'violet'},
17:{'name':'Cl','color':'brown'},
18:{'name':'Ar','color':'grey'},
19:{'name':'K','color':'grey'},
20:{'name':'Ca','color':'grey'},
21:{'name':'Sc','color':'grey'},
22:{'name':'Ti','color':'grey'},
23:{'name':'V','color':'grey'},
24:{'name':'Cr','color':'grey'},
25:{'name':'Mn','color':'grey'},
26:{'name':'Fe','color':'grey'},
27:{'name':'Co','color':'grey'},
28:{'name':'Ni','color':'grey'},
29:{'name':'Cu','color':'grey'},
30:{'name':'Zn','color':'grey'},
31:{'name':'Ga','color':'grey'},
32:{'name':'Ge','color':'grey'},
33:{'name':'As','color':'grey'},
34:{'name':'Se','color':'grey'},
35:{'name':'Br','color':'grey'},
36:{'name':'Kr','color':'grey'},
37:{'name':'Rb','color':'grey'},
38:{'name':'Sr','color':'grey'},
39:{'name':'Y','color':'grey'},
40:{'name':'Zr','color':'grey'},
41:{'name':'Nb','color':'grey'},
42:{'name':'Mo','color':'grey'},
43:{'name':'Tc','color':'grey'},
44:{'name':'Ru','color':'grey'},
45:{'name':'Rh','color':'grey'},
46:{'name':'Pd','color':'grey'},
47:{'name':'Ag','color':'grey'},
48:{'name':'Cd','color':'grey'},
49:{'name':'In','color':'grey'},
50:{'name':'Sn','color':'grey'},
51:{'name':'Sb','color':'grey'},
52:{'name':'Te','color':'grey'},
53:{'name':'I','color':'grey'},
54:{'name':'Xe','color':'grey'},
55:{'name':'Cs','color':'grey'},
56:{'name':'Ba','color':'grey'},
57:{'name':'Hf','color':'grey'},
58:{'name':'Ta','color':'grey'},
59:{'name':'W','color':'grey'},
60:{'name':'Re','color':'grey'},
61:{'name':'Os','color':'grey'},
62:{'name':'Ir','color':'grey'},
63:{'name':'Pt','color':'grey'},
64:{'name':'Au','color':'grey'},
65:{'name':'Hg','color':'grey'},
66:{'name':'Tl','color':'grey'},
67:{'name':'Pb','color':'grey'},
68:{'name':'Bi','color':'grey'},
69:{'name':'Po','color':'grey'},
70:{'name':'At','color':'grey'},
71:{'name':'Rn','color':'grey'},
72:{'name':'Fr','color':'grey'},
73:{'name':'Ra','color':'grey'},
74:{'name':'Pt','color':'grey'},
75:{'name':'Ac','color':'grey'},
76:{'name':'La','color':'grey'},
77:{'name':'U','color':'grey'},
78:{'name':'Sm','color':'grey'},
79:{'name':'Ce','color':'grey'},
80:{'name':'Nd','color':'grey'},
81:{'name':'Eu','color':'grey'},
82:{'name':'Gd','color':'grey'},
83:{'name':'Dy','color':'grey'},
84:{'name':'Er','color':'grey'},
85:{'name':'Rh','color':'grey'}}

#%
# copyright: https://github.com/tpoisot/nxfa2
def forceatlas2_layout(G, iterations=10, linlog=False, pos=None, nohubs=False,
                       kr=0.001, k=None, dim=2):
    """
    Options values are
    g                The graph to layout
    iterations       Number of iterations to do
    linlog           Whether to use linear or log repulsion
    random_init      Start with a random position
                     If false, start with FR
    avoidoverlap     Whether to avoid overlap of points
    degreebased      Degree based repulsion
    """
    # We add attributes to store the current and previous convergence speed
    for n in G:
        G.node[n]['prevcs'] = 0
        G.node[n]['currcs'] = 0
        # To numpy matrix
    # This comes from the spares FR layout in nx
    A = nx.to_scipy_sparse_matrix(G, dtype='f')
    nnodes, _ = A.shape

    try:
        A = A.tolil()
    except Exception as e:
        A = (coo_matrix(A)).tolil()
    if pos is None:
        pos = np.asarray(np.random.random((nnodes, dim)), dtype=A.dtype)
    else:
        pos = pos.astype(A.dtype)
    if k is None:
        k = np.sqrt(1.0 / nnodes)
        # Iterations
    # the initial "temperature" is about .1 of domain area (=1x1)
    # this is the largest step allowed in the dynamics.
    t = 0.1
    # simple cooling scheme.
    # linearly step down by dt on each iteration so last iteration is size dt.
    dt = t / float(iterations + 1)
    displacement = np.zeros((dim, nnodes))
    for iteration in range(iterations):
        displacement *= 0
        # loop over rows
        for i in range(A.shape[0]):
            # difference between this row's node position and all others
            delta = (pos[i] - pos).T
            # distance between points
            distance = np.sqrt((delta ** 2).sum(axis=0))
            # enforce minimum distance of 0.01
            distance = np.where(distance < 0.01, 0.01, distance)
            # the adjacency matrix row
            Ai = np.asarray(A.getrowview(i).toarray())
            # displacement "force"
            Dist = k * k / distance ** 2
            if nohubs:
                Dist = Dist / float(Ai.sum(axis=1) + 1)
            if linlog:
                Dist = np.log(Dist + 1)
            displacement[:, i] += \
                (delta * (Dist - Ai * distance / k)).sum(axis=1)
            # update positions
        length = np.sqrt((displacement ** 2).sum(axis=0))
        length = np.where(length < 0.01, 0.01, length)
        pos += (displacement * t / length).T
        # cool temperature
        t -= dt
        # Return the layout
    return dict(list(zip(G, pos)))

#%
def draw_isomorphism(G1,G2,layout=None):
    plt.clf()
    plt.axis('off')
    
    em = nx.algorithms.isomorphism.categorical_edge_match('weight', 1)
    nm = nx.algorithms.isomorphism.categorical_node_match('label', 'C')
    GM = nx.algorithms.isomorphism.GraphMatcher(G1,G2,node_match=nm,edge_match=em)
    
    if GM.subgraph_is_isomorphic():
        nodes_isomorphic=sorted(GM.mapping.keys())
        labels_isomorphic={v: nx.get_node_attributes(G1,'label')[v] for v in nodes_isomorphic}
        colors_isomorphic=[nx.get_node_attributes(G1,'color')[v] for v in nodes_isomorphic]
        ##nodes_normal=[v for v in G1.nodes() if not (v in nodes_isomorphic)]
        #edges_normal_single=[(u,v) for (u,v,d) in G1.edges(data=True) if (d['weight'] == 1) and not (u in nodes_isomorphic) and not (v in nodes_isomorphic)]
        #edges_normal_double=[(u,v) for (u,v,d) in G1.edges(data=True) if (d['weight'] == 2) and not (u in nodes_isomorphic) and not (v in nodes_isomorphic)] 
        edges_isomorphic_single=[(u,v) for (u,v,d) in G1.edges(data=True) if (d['weight'] == 1) and (u in nodes_isomorphic) and (v in nodes_isomorphic)]
        edges_isomorphic_double=[(u,v) for (u,v,d) in G1.edges(data=True) if (d['weight'] == 2) and (u in nodes_isomorphic) and (v in nodes_isomorphic)]
        
        edges_single1=[(u,v) for (u,v,d) in G1.edges(data=True) if d['weight'] == 1]
        edges_double1=[(u,v) for (u,v,d) in G1.edges(data=True) if d['weight'] == 2]
        #edges_single2=[(u,v) for (u,v,d) in G2.edges(data=True) if d['weight'] == 1]
        #edges_double2=[(u,v) for (u,v,d) in G2.edges(data=True) if d['weight'] == 2]
    
        if layout==None:
            #pos=nx.spring_layout(G1)
            pos1=forceatlas2_layout(G1,iterations=350,k=1,nohubs=True)
        else:
            pos1=layout
            
        # draw main graph (G1):
        nx.draw_networkx(G1,pos1,with_labels=True,labels=nx.get_node_attributes(G1,'label'),node_color=list(nx.get_node_attributes(G1,'color').values()),node_size=500,font_size=16,width=2,edgelist=edges_single1,alpha=0.2,style='dotted')
        nx.draw_networkx_edges(G1,pos1,edgelist=edges_double1,width=6,alpha=0.2,style='dotted')
        
        # draw isomorphic subgraph (G2):
        nx.draw_networkx(G1,pos1,nodelist=nodes_isomorphic,with_labels=True,labels=labels_isomorphic,node_color=colors_isomorphic,node_size=500,font_size=16,width=2,edgelist=edges_isomorphic_single)
        nx.draw_networkx_edges(G1,pos1,edgelist=edges_isomorphic_double,width=6)
        
        return pos1
    else:
        return layout
#%
def majority_vote_prediction(clf_list,test_X):
        
    test_results = np.zeros((np.size(test_X,-2),len(clf_list)),bool)
    for i in range(len(clf_list)):
	#This actually predicts all input data in one go
        test_results[:,i] = clf_list[i].predict(test_X[i])
	#If I want to catch the decision path, this has to happen here
        node_indicator = clf_list[i].decision_path(test_X[i])
    feature = clf_list[i].tree_.feature
    threshold = clf_list[i].tree_.threshold
    leave_id = clf_list[i].apply(test_X[i])
    for sample_id in range(len(test_X[i])):
        node_index = node_indicator.indices[node_indicator.indptr[sample_id]: node_indicator.indptr[sample_id + 1]]
        print('Rules used in tree %d to predict sample %s: ' % (i,sample_id))
        for node_id in node_index:
            if leave_id[sample_id] == node_id:
                continue

            if (test_X[i, sample_id, feature[node_id]] <= threshold[node_id]):
                threshold_sign = "<="
            else:
                threshold_sign = ">"

            if test_X[i, sample_id, feature[node_id]] == True and threshold_sign == ">":
                print("decision id node %s : (test_X[%s, %d] (= %s) %s %s)"
                    % (node_id,
                    sample_id,
                    feature[node_id],
                    test_X[i, sample_id, feature[node_id]],
                    threshold_sign,
                    threshold[node_id]))
		
    return np.array([bool(round(np.mean(el))) for el in test_results])
#%
def isomorph_test(G1,G2):
    em = nx.algorithms.isomorphism.categorical_edge_match('weight', 1)
    nm = nx.algorithms.isomorphism.categorical_node_match('label', 'C')
    GM = nx.algorithms.isomorphism.GraphMatcher(G1,G2,node_match=nm,edge_match=em) #G1-main graph, G2-subgraph
    
    return GM.subgraph_is_isomorphic()
#%  
def learn_classifier(num_molecules,num_substructures,num_samplings):  
    # load training data:
    training_data_X = np.load('classifier/training_data_X_1.npy')
    train_Y = np.load('classifier/training_data_Y.npy')
    #training_data_num = np.load('training_data_num_'+str(j+1)+'.npy')
    
    # initialize train and test sets:
    ## frequent substructures only:
    train_X = np.zeros((num_molecules*2,num_substructures),bool)
      
    ## numerical features only:
    #train_X = np.zeros((num_molecules*2,total_num_features),float)
    
    ## frequent substructures and numerical features:
    #train_X = np.zeros((num_molecules*2,num_substructures+total_num_features),float)
    
    # list of classifiers initialization:
    clf_list = []
    
    for i in range(num_samplings):
        print('learning',round(((i+1)/float(num_samplings))*100,2),'%')
        
        # define train and test data: 
        ## frequent substructures only:
        train_X = training_data_X[i]
        
        ## numerical features only:
        #train_X = training_data_num[i]
        
        ## frequent substructures and numerical features:
        #train_X[:,0:num_substructures] = training_data_X[i]
        #train_X[:,num_substructures:num_substructures+total_num_features] = training_data_num[i]
        
        # define classifier:
        clf = tree.DecisionTreeClassifier(criterion='entropy')
        
        #clf = tree.DecisionTreeClassifier()
        #clf = RandomForestClassifier()
        #clf = RandomForestClassifier(max_depth=4)
        #clf = svm.SVC(gamma=0.001,C=100)
        #clf = svm.SVC()
        #clf = GaussianNB()
        #clf = linear_model.LogisticRegression()
        #clf = SGDClassifier()
        
        # perform learning:
        clf = clf.fit(train_X,train_Y)
             
        # save classifier:
        clf_list.append(clf)
        
    return clf_list
#%
def open_dataset(input_file):
    #f = open('input/'+input_file+'.gsp','r')
    f = open(input_file,'r')
    M_list = []
    
    for line in f:
        try:
            if '#' in line:
                G1 = nx.Graph()
                line = next(f)
                while 'v' in line:
                    data = line.split(' ')
                    node = np.int(data[1])
                    element = np.int(data[2])
                    G1.add_node(node,label=CHEM_EL[element]['name'],color=CHEM_EL[element]['color'])
                    line = next(f)
                while 'e' in line:
                    data = line.split(' ')
                    vertice1 = np.int(data[1])
                    vertice2 = np.int(data[2])
                    G1.add_edge(vertice1,vertice2,weight=np.int(data[3]))
                    line = next(f)
                M_list.append(G1)
        except ValueError:
            print("Invalid input:", line)
    f.close()
    
    return M_list
#%%   
def encode_dataset(M_list,num_substructures,num_samplings):    
    # number of molecules for classification:
    num_molecules_classify = len(M_list)
    
    # dimensions of input data:
    pred_X = np.zeros((num_samplings,num_molecules_classify,num_substructures),bool)
    
    for i in range(num_samplings):
        UpdateProgressBar(round(((i+1)/float(num_samplings))*100,2))
        #print 'encoding',round(((i+1)/float(num_samplings))*100,2),'%'
    
        # Open top discriminative substructures for encoding the input data:
        f = open('classifier/top_sg_1_'+str(i+1)+'.gsp','r')
        SG_list = []
        
        for line in f:
            try:
                if '#' in line:
                    G2 = nx.Graph()
                    #labels2 = {}
                    #node_colors2 = []
                    line = next(f)
                    while 'v' in line:
                        data = line.split(' ')
                        node = np.int(data[1])
                        element = np.int(data[2])
                        #labels2[node] = CHEM_EL[element]['name']
                        #node_colors2.append(CHEM_EL[element]['color'])
                        #G2.add_node(node,label=labels2[node])
                        G2.add_node(node,label=CHEM_EL[element]['name'])
                        line = next(f)
                    while 'e' in line:
                        data = line.split(' ')
                        vertice1 = np.int(data[1])
                        vertice2 = np.int(data[2])
                        G2.add_edge(vertice1,vertice2,weight=np.int(data[3]))
                        line = next(f)
                    SG_list.append(G2)
            except ValueError:
                print("Invalid input:", line)
        f.close()
        
        # Encode the input data:
        for i_molecule in range(num_molecules_classify):
            for i_substructure in range(len(SG_list)):
                pred_X[i][i_molecule][i_substructure] = isomorph_test(M_list[i_molecule],SG_list[i_substructure])
                
    return pred_X
#%
def molecule_stat(test_results,test_X,molecule_id):
    # list of subgraphs which exist in molecule:
    molecule_subgraphs_str = []
    buff_str = ''
    
    for i in range(num_samplings):
        if test_results[molecule_id] == True: # if molecule classified as frequent hitter
            # list of indexes of subgraphs which exist in molecule
            subgraphs_list = [l for l, k in enumerate(test_X[i,molecule_id]) if k == True]
            
            # load the subgraphs
            j = 0 # open fold 1
            f = open('classifier/top_sg_'+str(j+1)+'_'+str(i+1)+'.gsp','r')
            buff_str_list = []    
            for line in f:
                try:
                    if '#' in line:
                        while not (line == '\n'):
                            line = next(f)
                            buff_str += line
                        buff_str_list.append(buff_str)
                        buff_str = ''
                except ValueError:
                    print("Invalid input:", line)
            f.close()
            
            # form the list of subgraphs
            for subgraph in subgraphs_list:
                molecule_subgraphs_str.append(buff_str_list[subgraph])
    #%
    # list of unique subgraphs which exist in molecule:
    f_out = open('subgraphs.dat', "w")
    subgraphs_frequencies = []
    
    counter = 0
    for graph,freq in collections.Counter(molecule_subgraphs_str).most_common():
        f_out.write('#'+str(counter+1)+' th subgraph appears '+str(freq)+' times\n')
        f_out.write(graph)
        subgraphs_frequencies.append(freq)
        counter += 1
        
    f_out.close()
    #%
    f = open('subgraphs.dat','r')
    SG_list = []
    
    for line in f:
        try:
            if '#' in line:
                G3 = nx.Graph()
                line = next(f)
                while 'v' in line:
                    data = line.split(' ')
                    node = np.int(data[1])
                    element = np.int(data[2])
                    G3.add_node(node,label=CHEM_EL[element]['name'],color=CHEM_EL[element]['color'])
                    line = next(f)
                while 'e' in line:
                    data = line.split(' ')
                    vertice1 = np.int(data[1])
                    vertice2 = np.int(data[2])
                    G3.add_edge(vertice1,vertice2,weight=np.int(data[3]))
                    line = next(f)
                SG_list.append(G3)
        except ValueError:
            print ("Invalid input:", line)
    f.close()
    
    return SG_list,subgraphs_frequencies
#%%
# model parameters:
num_molecules = 812
num_substructures = 100
#num_folds = 10
num_samplings = 200
#total_num_features = 192

# learn classifier:
clf_list = learn_classifier(num_molecules,num_substructures,num_samplings)
#%%
# open dataset for classification:
#input_file = 'freq_test_1'#'S4'
#M_list = open_dataset(input_file)
#%%
# encode dataset:
#pred_X = encode_dataset(M_list,num_substructures,num_samplings)
#%%
# perform prediction:
#pred_results = np.zeros(len(M_list),bool)
#pred_results = majority_vote_prediction(clf_list,pred_X)
#%%
#from Tkinter import *
import tkinter.filedialog
import tkinter.ttk

G=nx.Graph()
POS = None #POS = draw_isomorphism(M_list[molecule_id],G)

fig = plt.gcf()

master = Tk()

frameTopLeft = Frame(master)
frameTopLeft.grid(row=1,column=1)

frameTopRight = Frame(master)
frameTopRight.grid(row=1,column=2)

topLabelM = Label(master,text='Choose molecule:')
topEntryM = Entry(master,width=10)
topLabelM.grid(row=2,column=1)

topLabelG = Label(master,text='Choose subgraph:')
topEntryG = Entry(master,width=10)
topLabelG.grid(row=2,column=2)

frameLeft = Frame(master)
frameLeft.grid(row=3,column=1)

frameRight = Frame(master)
frameRight.grid(row=3,column=2)

bottomLeftLabel = Label(master,text='')
bottomLeftEntry = Entry(master,width=10)
bottomLeftLabel.grid(row=4,column=1)

bottomRightLabel = Label(master,text='')
bottomRightEntry = Entry(master,width=10)
bottomRightLabel.grid(row=4,column=2)

def LoadFile(ev):
    global f_name
    f_name = tkinter.filedialog.Open(master,filetypes = [('*.gsp files', '.gsp')]).show()
    if f_name == '':
        return
    
def EncodeFile(ev):
    global M_dataset
    global dataset_X
    global pred_res
    M_dataset = open_dataset(f_name)
    for item in range(len(M_dataset)):
        listboxMolecules.insert(END,'molecule '+str(item+1))
    dataset_X = encode_dataset(M_dataset,num_substructures,num_samplings)
    
    pred_res = np.zeros(len(M_dataset),bool)
    pred_res = majority_vote_prediction(clf_list,dataset_X)
    bottomLeftLabel['text'] = ''
    
def Quit(ev):
    global master
    plt.close("all")
    master.destroy()
   
def MoleculeSelect(evt):
    global POS
    global graphs_list
    global graphs_freq
    global molecule_index
    
    molecule_index = listboxMolecules.curselection()[0]
    
    if pred_res[molecule_index]:
        bottomLeftLabel['text'] = "Result: frequent"
    else:
        bottomLeftLabel['text'] = "Result: none-frequent"
        
    POS = draw_isomorphism(M_dataset[molecule_index],G)
    graphs_list,graphs_freq = molecule_stat(pred_res,dataset_X,molecule_index)
    
    listboxGraphs.delete(0, END)
    for item in range(len(graphs_list)):
        listboxGraphs.insert(END,'subgraph '+str(item+1))
    bottomRightLabel['text'] = ''

def GraphSelect(evt):
    graph_index = listboxGraphs.curselection()[0]
    
    bottomRightLabel['text'] = "Frequency: "+str(graphs_freq[graph_index])
    
    pos = draw_isomorphism(M_dataset[molecule_index],graphs_list[graph_index],POS)
    
def UpdateProgressBar(progress):
    encodingBar["value"] = progress
    encodingBar.update()
    bottomLeftLabel['text'] = "Encoding: "+str(progress)+"%"
    #print 'encoding',progress,'%'
    
loadBtn = Button(frameTopLeft, text = 'Open dataset')
loadBtn.bind("<ButtonRelease-1>", LoadFile)
loadBtn.pack(side = 'left')

encodeBtn = Button(frameTopLeft, text = 'Encoding')
encodeBtn.bind("<ButtonRelease-1>", EncodeFile)
encodeBtn.pack(side = 'right')

encodingBar = tkinter.ttk.Progressbar(frameTopRight,orient ="horizontal",length = 150,mode ="determinate")
encodingBar.pack(side = 'left')
encodingBar["maximum"] = 100

quitBtn = Button(frameTopRight,text = 'Quit')
quitBtn.bind("<Button-1>",Quit)
quitBtn.pack(side = 'right')

listboxMolecules = Listbox(frameLeft,selectmode=SINGLE,height=20)
listboxMolecules.bind('<<ListboxSelect>>',MoleculeSelect)
listboxMolecules.pack(side="left")

scrollbarLeft = Scrollbar(frameLeft,orient="vertical")
scrollbarLeft.config(command=listboxMolecules.yview)
scrollbarLeft.pack(side="right",fill="y")
listboxMolecules.config(yscrollcommand=scrollbarLeft.set)

listboxGraphs = Listbox(frameRight,selectmode=SINGLE,height=20)
listboxGraphs.bind('<<ListboxSelect>>',GraphSelect)
listboxGraphs.pack(side="left")

scrollbarRight = Scrollbar(frameRight,orient="vertical")
scrollbarRight.config(command=listboxGraphs.yview)
scrollbarRight.pack(side="right",fill="y")
listboxGraphs.config(yscrollcommand=scrollbarRight.set)

master.mainloop()
#%%
