#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import collections

import networkx as nx
import numpy as np
import os
import logging
import time
import math
import sys

from datetime import datetime
from sklearn import tree

# from openbabel import openbabel
# from openbabel import pybel 

import openbabel
import pybel

CHEM_EL = {
    1: {'name': 'H', 'color': 'yellow'},
    2: {'name': 'He', 'color': 'grey'},
    3: {'name': 'Li', 'color': 'grey'},
    4: {'name': 'Be', 'color': 'grey'},
    5: {'name': 'B', 'color': 'grey'},
    6: {'name': 'C', 'color': 'orange'},
    7: {'name': 'N', 'color': 'red'},
    8: {'name': 'O', 'color': 'green'},
    9: {'name': 'F', 'color': 'blue'},
    10: {'name': 'Ne', 'color': 'grey'},
    11: {'name': 'Na', 'color': 'grey'},
    12: {'name': 'Mg', 'color': 'grey'},
    13: {'name': 'Al', 'color': 'grey'},
    14: {'name': 'Si', 'color': 'grey'},
    15: {'name': 'P', 'color': 'grey'},
    16: {'name': 'S', 'color': 'violet'},
    17: {'name': 'Cl', 'color': 'brown'},
    18: {'name': 'Ar', 'color': 'grey'},
    19: {'name': 'K', 'color': 'grey'},
    20: {'name': 'Ca', 'color': 'grey'},
    21: {'name': 'Sc', 'color': 'grey'},
    22: {'name': 'Ti', 'color': 'grey'},
    23: {'name': 'V', 'color': 'grey'},
    24: {'name': 'Cr', 'color': 'grey'},
    25: {'name': 'Mn', 'color': 'grey'},
    26: {'name': 'Fe', 'color': 'grey'},
    27: {'name': 'Co', 'color': 'grey'},
    28: {'name': 'Ni', 'color': 'grey'},
    29: {'name': 'Cu', 'color': 'grey'},
    30: {'name': 'Zn', 'color': 'grey'},
    31: {'name': 'Ga', 'color': 'grey'},
    32: {'name': 'Ge', 'color': 'grey'},
    33: {'name': 'As', 'color': 'grey'},
    34: {'name': 'Se', 'color': 'grey'},
    35: {'name': 'Br', 'color': 'grey'},
    36: {'name': 'Kr', 'color': 'grey'},
    37: {'name': 'Rb', 'color': 'grey'},
    38: {'name': 'Sr', 'color': 'grey'},
    39: {'name': 'Y', 'color': 'grey'},
    40: {'name': 'Zr', 'color': 'grey'},
    41: {'name': 'Nb', 'color': 'grey'},
    42: {'name': 'Mo', 'color': 'grey'},
    43: {'name': 'Tc', 'color': 'grey'},
    44: {'name': 'Ru', 'color': 'grey'},
    45: {'name': 'Rh', 'color': 'grey'},
    46: {'name': 'Pd', 'color': 'grey'},
    47: {'name': 'Ag', 'color': 'grey'},
    48: {'name': 'Cd', 'color': 'grey'},
    49: {'name': 'In', 'color': 'grey'},
    50: {'name': 'Sn', 'color': 'grey'},
    51: {'name': 'Sb', 'color': 'grey'},
    52: {'name': 'Te', 'color': 'grey'},
    53: {'name': 'I', 'color': 'grey'},
    54: {'name': 'Xe', 'color': 'grey'},
    55: {'name': 'Cs', 'color': 'grey'},
    56: {'name': 'Ba', 'color': 'grey'},
    57: {'name': 'Hf', 'color': 'grey'},
    58: {'name': 'Ta', 'color': 'grey'},
    59: {'name': 'W', 'color': 'grey'},
    60: {'name': 'Re', 'color': 'grey'},
    61: {'name': 'Os', 'color': 'grey'},
    62: {'name': 'Ir', 'color': 'grey'},
    63: {'name': 'Pt', 'color': 'grey'},
    64: {'name': 'Au', 'color': 'grey'},
    65: {'name': 'Hg', 'color': 'grey'},
    66: {'name': 'Tl', 'color': 'grey'},
    67: {'name': 'Pb', 'color': 'grey'},
    68: {'name': 'Bi', 'color': 'grey'},
    69: {'name': 'Po', 'color': 'grey'},
    70: {'name': 'At', 'color': 'grey'},
    71: {'name': 'Rn', 'color': 'grey'},
    72: {'name': 'Fr', 'color': 'grey'},
    73: {'name': 'Ra', 'color': 'grey'},
    74: {'name': 'Pt', 'color': 'grey'},
    75: {'name': 'Ac', 'color': 'grey'},
    76: {'name': 'La', 'color': 'grey'},
    77: {'name': 'U', 'color': 'grey'},
    78: {'name': 'Sm', 'color': 'grey'},
    79: {'name': 'Ce', 'color': 'grey'},
    80: {'name': 'Nd', 'color': 'grey'},
    81: {'name': 'Eu', 'color': 'grey'},
    82: {'name': 'Gd', 'color': 'grey'},
    83: {'name': 'Dy', 'color': 'grey'},
    84: {'name': 'Er', 'color': 'grey'},
    85: {'name': 'Rh', 'color': 'grey'}
}

CLASSIFIER_DIRECTORY_PATH = "classifier"
M_FILE = "demo2.sdf"

def learn_classifier(num_samplings):

    # num_samplings = 200
    logging.debug("CLASSIFIER_DIRECTORY_PATH is [%s]" % (CLASSIFIER_DIRECTORY_PATH))

    # load training data:
    training_data_X = np.load(CLASSIFIER_DIRECTORY_PATH + '/training_data_X.npy')
    train_Y = np.load(CLASSIFIER_DIRECTORY_PATH  + '/training_data_Y.npy')

    # list of classifiers initialization:
    clf_list = []

    trees_dir_path = "trees-" + str(time.strftime("%Y%m%d-%H%M%S"))

    for i in range(num_samplings):

        logging.info('learning %s %%' %(round(((i + 1)/float(num_samplings)) * 100, 2)))
        print('learning %s %%' %(round(((i + 1)/float(num_samplings)) * 100, 2)))

        train_X = training_data_X[i]

        if i == 0:
            
            logging.debug("*** trains ***")
            logging.debug("train_X.ndim is [%s]" % (train_X.ndim))
            logging.debug("train_X.shape is [%s]" % (str(train_X.shape)))
            logging.debug("train_Y.ndim is [%s]" % (train_Y.ndim))
            logging.debug("train_Y.shape is [%s]" % (str(train_Y.shape)))
            logging.debug("*** end of trains ***")

            #
            # output for demo2.sdf
            #
            # train_X.ndim is [ 2 ]
            # train_Y.ndim is [ 1 ]
            #

        # define classifier:
        clf = tree.DecisionTreeClassifier(criterion='entropy', min_samples_leaf=17*3)

        # perform learning:
        clf = clf.fit(train_X, train_Y)

        # if logging.getLogger().isEnabledFor(logging.DEBUG):

        if not os.path.isdir(trees_dir_path):

            os.mkdir(trees_dir_path)

        tree.export_graphviz(clf, out_file=trees_dir_path + os.path.sep + str(i) + ".dot")

        # save classifier:
        clf_list.append(clf)

    return clf_list


def open_dataset(input_file):
    # open molecules:
    molecules = pybel.readfile("sdf", input_file)

    M_list = []
    M_list_pybel = []

    for molecule in molecules:

        G1 = nx.Graph()
        # add vertices:
        for atom in molecule:
            node = atom.idx-1
            element = atom.atomicnum
            G1.add_node(node, label=CHEM_EL[element]['name'], color=CHEM_EL[element]['color'])
        # add edges:
        for bond in openbabel.OBMolBondIter(molecule.OBMol):
            vertice1 = bond.GetBeginAtom().GetIdx()-1
            vertice2 = bond.GetEndAtom().GetIdx()-1
            G1.add_edge(vertice1, vertice2, weight=bond.GetBondOrder())
        # memorize molecule:
        M_list.append(G1)
        M_list_pybel.append(molecule)

    return M_list, M_list_pybel


def encode_dataset(M_list, num_substructures, num_samplings):
    """

    Encodes, for each molecule, the subgraphs associated to the classifiers.

    Args:

        M_list (list):
        num_substructures (int): 100
        num_samplings (int): 200

    Returns:

        A numpy array, with 3 dimensions : classifiers, molecules,
        substructures.

        Eg. 200 / 4 / 100.

        For each record, we know if the substructure (subgraph) is
        iso-morphic or not with the molecule.

        If it is iso-morph, this means, the sub-graph is part of the molecule.

        It has to be displayed.

    """

    logging.debug("CLASSIFIER_DIRECTORY_PATH is [%s]" % (CLASSIFIER_DIRECTORY_PATH))

    begin = time.time()

    # number of molecules for classification:
    num_molecules_classify = len(M_list)

    # dimensions of input data:
    pred_X = np.zeros((num_samplings, num_molecules_classify, num_substructures), bool)

    for i in range(num_samplings):
        # UpdateProgressBar(round(((i+1)/float(num_samplings))*100,2))
        logging.info('encoding %s %%' % (round(((i+1)/float(num_samplings))*100, 2)))
        print('encoding %s %%' % (round(((i+1)/float(num_samplings))*100, 2)))

        # Open top discriminative substructures for encoding the input data:
        f = open(CLASSIFIER_DIRECTORY_PATH + '/top_sg_'+str(i+1)+'.gsp', 'r')
        SG_list = []

        for line in f:
            try:
                if '#' in line:
                    G2 = nx.Graph()
                    # labels2 = {}
                    # node_colors2 = []
                    line = next(f)
                    while 'v' in line:
                        data = line.split(' ')
                        node = np.int(data[1])
                        element = np.int(data[2])
                        # labels2[node] = CHEM_EL[element]['name']
                        # node_colors2.append(CHEM_EL[element]['color'])
                        # G2.add_node(node,label=labels2[node])
                        G2.add_node(node, label=CHEM_EL[element]['name'])
                        line = next(f)
                    while 'e' in line:
                        data = line.split(' ')
                        vertice1 = np.int(data[1])
                        vertice2 = np.int(data[2])
                        G2.add_edge(vertice1, vertice2, weight=np.int(data[3]))
                        line = next(f)
                    SG_list.append(G2)
            except ValueError:
                print("Invalid input:", line)
        f.close()

        # Encode the input data:
        for i_molecule in range(num_molecules_classify):
            for i_substructure in range(len(SG_list)):
                pred_X[i][i_molecule][i_substructure] = isomorph_test(M_list[i_molecule], SG_list[i_substructure])

    end = time.time()

    logging.info("time is [%s]" % math.ceil(end - begin))

    return pred_X


def isomorph_test(G1, G2):

    em = nx.algorithms.isomorphism.categorical_edge_match('weight', 1)
    nm = nx.algorithms.isomorphism.categorical_node_match('label', 'C')
    GM = nx.algorithms.isomorphism.GraphMatcher(G1, G2, node_match=nm, edge_match=em)  # G1-main graph, G2-subgraph

    return GM.subgraph_is_isomorphic()


def majority_vote_prediction(clf_list, test_X):
    """This function does the prediction vote for the given classifier list.

Args:
    clf_list (list): the classifier list.
    test_X (numpy.ndarray): a numpy array, with, for each molecule, an array
    of booleans (the molecule encoded) and a boolean if it is iso-morphic or
    not with the given substructure.

Returns:
    A numpy array with, for each molecule (index), True if it is frequent or
    False if it is non-frequent.

    Only frequent molecules have to be displayed.
    """

    test_results = np.zeros((np.size(test_X, -2), len(clf_list)), bool)

    for i in range(len(clf_list)):

        logging.debug("test_X[i].ndim is [%s]" % (test_X[i].ndim))
        logging.debug("test_X[i].shape is [%s]" % (str(test_X[i].shape)))
        logging.debug("test_X[i] is [%s]" % (test_X[i]))

        #
        # output for demo2.sdf :
        #
        # test_X[i].ndim is [ 2 ]
        # test_X[i].shape is [ (4, 100) ]
        #

        test_results[:, i] = clf_list[i].predict(test_X[i])

    result = np.array([bool(round(np.mean(el))) for el in test_results])

    logging.debug("result.ndim is [%s]" % (result.ndim))
    logging.debug("result.shape is [%s]" % (result.shape))
    logging.debug("result is [%s]" % (result))

    return result


def molecule_stat(clf_list, molecule_id, dataset_X, num_samplings, num_substructures):

    logging.debug("CLASSIFIER_DIRECTORY_PATH is [%s]" % (CLASSIFIER_DIRECTORY_PATH))

    molecule_subgraphs_str = []

    for classifier_id in range(num_samplings):

        logging.info("**********************************************")
        logging.info("classifier_id is [%s]" %(classifier_id))

        test_X = dataset_X[classifier_id, molecule_id].reshape(1, num_substructures)
            
        # logging.info("classifier_id is [%s]" % (classifier_id))
        # logging.info("test_X.ndim is [%s]" % (test_X.ndim))
        # logging.info("test_X.shape is [%s]" % (str(test_X.shape)))
        # logging.info("test_X is [%s]" % (test_X)) 

        #
        # output for demo2.sdf  :
        #
        # test_X.ndim[1]
        # (100,)
        #

        subgraphs_list = []

        for l, k in enumerate(test_X[0]):

            if k:

                subgraphs_list.append(l)

        logging.info("subgraphs_list is [%s]" % (subgraphs_list))

        subgraphs_list = []

        prediction = clf_list[classifier_id].predict(test_X)

        if not prediction[0]:

            logging.info("prediction is false")
        
        else:

            decision_path = clf_list[classifier_id].decision_path(test_X)

            logging.debug("decision_path is [%s]" % (decision_path))

            feature = clf_list[classifier_id].tree_.feature
            threshold = clf_list[classifier_id].tree_.threshold

            leave_id = clf_list[classifier_id].apply(test_X)

            for sample_id in range(len(test_X)):

                node_index = decision_path.indices[decision_path.indptr[sample_id]: decision_path.indptr[sample_id + 1]]

                logging.debug("sample_id is [%s]" % (sample_id))
                logging.debug("node_index is [%s]" % (node_index))
                logging.debug("feature is [%s]" % (feature))
                logging.debug("threshold is [%s]" % (threshold))
                
                logging.info("leave_id is [%s]" % (leave_id))
                logging.info('Rules used in tree %d to predict sample %s: ' % (classifier_id, sample_id))
                #  print('Rules used in tree %d to predict sample %s: ' % (classifier_id, sample_id))

                for node_id in node_index:

                    logging.info("node_id is [%s]" % (node_id))

                    if leave_id[sample_id] == node_id:
                        continue

                    if (test_X[sample_id, feature[node_id]] <= threshold[node_id]):
                        threshold_sign = "<="
                    else:
                        threshold_sign = ">"

                    if test_X[sample_id, feature[node_id]] and threshold_sign == ">":

                        logging.info("decision id node %s : (test_X[%s, %d] (= %s) %s %s)"
                            % (node_id,
                                sample_id,
                                feature[node_id],
                                test_X[sample_id, feature[node_id]],
                                threshold_sign,
                                threshold[node_id]))

                        subgraphs_list.append(feature[node_id])

        logging.info("new subgraphs_list is [%s]" % (str(subgraphs_list)))

        logging.debug("loading subgraphs from top_sg_%s.gsp" % (str(classifier_id + 1)))

        f = open(CLASSIFIER_DIRECTORY_PATH + '/top_sg_' + str(classifier_id + 1) + '.gsp', 'r')
        buff_str_list = []
        buff_str = ''
        for line in f:
            try:
                if '#' in line:
                    while ('#' in line) or ('v' in line) or ('e' in line):
                        line = next(f)
                        buff_str += line
                    # logging.debug("adding buff_str [%s] to buff_str_list" % (buff_str))
                    buff_str_list.append(buff_str)
                    buff_str = ''
            except ValueError:
                print("Invalid input:", line)
        f.close()

        # form the list of subgraphs
        for subgraph in subgraphs_list:
            logging.debug("subgraph [%s] is [%s]" % (subgraph, buff_str_list[subgraph]))
            molecule_subgraphs_str.append(buff_str_list[subgraph])

    top_sg_list = []
    subgraphs_frequencies = []

    for graph, freq in collections.Counter(molecule_subgraphs_str).most_common():
        top_sg_list.append('#\n'+graph)
        subgraphs_frequencies.append(freq)

    # SG_list = []
    Mol_list = []
    for line in top_sg_list:
        try:
            if '#' in line:
                line_lst = line.rstrip().split('\n')
                # G3 = nx.Graph()
                mol = openbabel.OBMol()
                for entry in line_lst:
                    if 'v' in entry:
                        data = entry.split(' ')
                        # node = np.int(data[1])
                        element = np.int(data[2])
                        # G3.add_node(node,label=CHEM_EL[element]['name'],color=CHEM_EL[element]['color'])
                        a = mol.NewAtom()
                        a.SetAtomicNum(element)
                        logging.debug("setting AtomicNum[%s]" % (element))
                        logging.debug("CHEM_EL[%s]['name'] is [%s]" % (element, CHEM_EL[element]['name']))
                    elif 'e' in entry:
                        data = entry.split(' ')
                        vertice1 = np.int(data[1])
                        vertice2 = np.int(data[2])
                        # G3.add_edge(vertice1,vertice2,weight=np.int(data[3]))
                        mol.AddBond(vertice1+1, vertice2+1, np.int(data[3]))
                # SG_list.append(G3)
                mol.AddHydrogens(True, False, 7.4)
                pybelMol = pybel.Molecule(mol)
                pybelMol.addh()
                logging.debug("adding pybelMol [%s]" % (pybelMol.formula))
                Mol_list.append(pybelMol)
        except ValueError:
            print("Invalid input:", line)

    return Mol_list, subgraphs_frequencies


if __name__ == "__main__":

    now = datetime.now()

    logFile = "prepep_ml-" + now.strftime("%Y%m%d-%H%M%S.%f") + ".log"

    logging.basicConfig(filename=logFile, level=logging.INFO)

    if len(sys.argv) == 3:
        if sys.argv[1] == "-d":
            CLASSIFIER_DIRECTORY_PATH = sys.argv[2]
        if sys.argv[1] == "-f":
            M_FILE = sys.argv[2]

    if len(sys.argv) == 5:
        if sys.argv[1] == "-d":
            CLASSIFIER_DIRECTORY_PATH = sys.argv[2]
            if sys.argv[3] == "-f":
                M_FILE = sys.argv[4]
        if sys.argv[1] == "-f":
            M_FILE = sys.argv[2]
            if sys.argv[3] == "-d":
                CLASSIFIER_DIRECTORY_PATH = sys.argv[4]

    encoded_training_data = np.load(CLASSIFIER_DIRECTORY_PATH + '/training_data_X.npy')
    num_samplings = encoded_training_data.shape[0]

    clf_list = learn_classifier(num_samplings)

    M_list, M_list_pybel = open_dataset(M_FILE)

    logging.debug("M_list length is [%s]" % (len(M_list)))
    logging.debug("M_list_pybel length is [%s]" % (len(M_list_pybel)))

    num_substructures = encoded_training_data.shape[2]

    dataset_X = encode_dataset(M_list, num_substructures, num_samplings)

    logging.debug("dataset_X.ndim is [%s]" % (dataset_X.ndim))
    logging.debug("dataset_X.shape is [%s]" % (str(dataset_X.shape)))

    #
    # output for demo2.sdf
    #
    # dataset_X.ndim is [ 3 ]
    # dataset_X.shape is [ (200, 4, 100) ]
    #

    pred_res = majority_vote_prediction(clf_list, dataset_X)

    logging.debug("pred_res.ndim is [%s]" % (pred_res.ndim))
    logging.debug("pred_res is [%s]" % (pred_res))

    #
    # output for demo2.sdf
    #
    # pred_res.ndim is [ 1 ]
    # pred_res is [ [False False  True  True] ]
    #

    for molecule_id in range(len(pred_res)):

        logging.info("processing molecule [%s]" % (molecule_id))

        formula = M_list_pybel[molecule_id].formula

        is_frequent = pred_res[molecule_id]

        if is_frequent:
            print("[%s] is frequent" % (formula))
            logging.info("[%s] is a frequent hitter" % (formula))
        else:
            print("[%s] is not frequent" % (formula))
            logging.info("[%s] is not a frequent hitter" % (formula))

        if is_frequent:

            Mol_list, subgraphs_frequencies = molecule_stat(clf_list, molecule_id, dataset_X, num_samplings, num_substructures)

            # logging.debug("Mol_list is [" + str(Mol_list) + "]")
            logging.info("Mol_list length is [%s]" % (len(Mol_list)))
            logging.info("subgraphs_frequencies is [%s]" % (subgraphs_frequencies))

            for Mol in Mol_list:
                logging.info(str(Mol.formula))

