import numpy as np
import sys
import random
import subprocess
import os
import networkx as nx
#%%

def isomorph_test(G1,G2):
    em = nx.algorithms.isomorphism.categorical_edge_match('weight', 1)
    nm = nx.algorithms.isomorphism.categorical_node_match('label', 'C')
    GM = nx.algorithms.isomorphism.GraphMatcher(G1,G2,node_match=nm,edge_match=em) #G1-main graph, G2-subgraph
    
    return GM.subgraph_is_isomorphic() 

def encode(infreqStump, freqFile, numSamplings, numSubgraphs, sampleSize):

    CHEM_EL = {1:{'name':'H','color':'yellow'},
    2:{'name':'He','color':'grey'},
    3:{'name':'Li','color':'grey'},
    4:{'name':'Be','color':'grey'},
    5:{'name':'B','color':'grey'},
    6:{'name':'C','color':'orange'},
    7:{'name':'N','color':'red'},
    8:{'name':'O','color':'green'},
    9:{'name':'F','color':'blue'},
    10:{'name':'Ne','color':'grey'},
    11:{'name':'Na','color':'grey'},
    12:{'name':'Mg','color':'grey'},
    13:{'name':'Al','color':'grey'},
    14:{'name':'Si','color':'grey'},
    15:{'name':'P','color':'grey'},
    16:{'name':'S','color':'violet'},
    17:{'name':'Cl','color':'brown'},
    18:{'name':'Ar','color':'grey'},
    19:{'name':'K','color':'grey'},
    20:{'name':'Ca','color':'grey'},
    21:{'name':'Sc','color':'grey'},
    22:{'name':'Ti','color':'grey'},
    23:{'name':'V','color':'grey'},
    24:{'name':'Cr','color':'grey'},
    25:{'name':'Mn','color':'grey'},
    26:{'name':'Fe','color':'grey'},
    27:{'name':'Co','color':'grey'},
    28:{'name':'Ni','color':'grey'},
    29:{'name':'Cu','color':'grey'},
    30:{'name':'Zn','color':'grey'},
    31:{'name':'Ga','color':'grey'},
    32:{'name':'Ge','color':'grey'},
    33:{'name':'As','color':'grey'},
    34:{'name':'Se','color':'grey'},
    35:{'name':'Br','color':'grey'},
    36:{'name':'Kr','color':'grey'},
    37:{'name':'Rb','color':'grey'},
    38:{'name':'Sr','color':'grey'},
    39:{'name':'Y','color':'grey'},
    40:{'name':'Zr','color':'grey'},
    41:{'name':'Nb','color':'grey'},
    42:{'name':'Mo','color':'grey'},
    43:{'name':'Tc','color':'grey'},
    44:{'name':'Ru','color':'grey'},
    45:{'name':'Rh','color':'grey'},
    46:{'name':'Pd','color':'grey'},
    47:{'name':'Ag','color':'grey'},
    48:{'name':'Cd','color':'grey'},
    49:{'name':'In','color':'grey'},
    50:{'name':'Sn','color':'grey'},
    51:{'name':'Sb','color':'grey'},
    52:{'name':'Te','color':'grey'},
    53:{'name':'I','color':'grey'},
    54:{'name':'Xe','color':'grey'},
    55:{'name':'Cs','color':'grey'},
    56:{'name':'Ba','color':'grey'},
    57:{'name':'Hf','color':'grey'},
    58:{'name':'Ta','color':'grey'},
    59:{'name':'W','color':'grey'},
    60:{'name':'Re','color':'grey'},
    61:{'name':'Os','color':'grey'},
    62:{'name':'Ir','color':'grey'},
    63:{'name':'Pt','color':'grey'},
    64:{'name':'Au','color':'grey'},
    65:{'name':'Hg','color':'grey'},
    66:{'name':'Tl','color':'grey'},
    67:{'name':'Pb','color':'grey'},
    68:{'name':'Bi','color':'grey'},
    69:{'name':'Po','color':'grey'},
    70:{'name':'At','color':'grey'},
    71:{'name':'Rn','color':'grey'},
    72:{'name':'Fr','color':'grey'},
    73:{'name':'Ra','color':'grey'},
    74:{'name':'Pt','color':'grey'},
    75:{'name':'Ac','color':'grey'},
    76:{'name':'La','color':'grey'},
    77:{'name':'U','color':'grey'},
    78:{'name':'Sm','color':'grey'},
    79:{'name':'Ce','color':'grey'},
    80:{'name':'Nd','color':'grey'},
    81:{'name':'Eu','color':'grey'},
    82:{'name':'Gd','color':'grey'},
    83:{'name':'Dy','color':'grey'},
    84:{'name':'Er','color':'grey'},
    85:{'name':'Rh','color':'grey'}}

    # dimensions of training datasets:
    training_data_X = np.zeros((numSamplings,sampleSize*2,numSubgraphs),bool)
    # classes of training dataset:
    training_data_Y = np.zeros((sampleSize*2),bool)
    training_data_Y[0:sampleSize] = True

    # frequent molecules for training:
    f = open(freqFile,'r')
    M_list_freq = []
    
    ##count = 0
    for line in f:
        try:
            if '#' in line:
                G1 = nx.Graph()
                #labels1 = {}
                #node_colors1 = []
                line = f.readline()
                while 'v' in line:
                    data = line.split(' ')
                    node = np.int(data[1])
                    element = np.int(data[2])
                    #labels1[node] = CHEM_EL[element]['name']
                    #node_colors1.append(CHEM_EL[element]['color'])
                    #G1.add_node(node,label=labels1[node])
                    G1.add_node(node,label=CHEM_EL[element]['name'])
                    line = f.readline()
                while 'e' in line:
                    data = line.split(' ')
                    vertice1 = np.int(data[1])
                    vertice2 = np.int(data[2])
                    G1.add_edge(vertice1,vertice2,weight=np.int(data[3]))
                    line = f.readline()
                M_list_freq.append(G1)
                ##if isomorph_test(G1,G2):
                ##    bin_matrix[count] = True
                ##count += 1
        except ValueError:
            print("Invalid input:"+line)
    f.close()

    # set of experiments inside one fold:
    print("Encoding sample", end='', flush=True)
    for i in range(numSamplings):
       
        print(" "+str(i+1), end='', flush=True)

        # sample of infrequent molecules for training:
        infreqFile = "sample_data/"+infreqStump+"_sample"+str(i+1)+".gsp"
        f = open(infreqFile,'r')
        M_list_infreq = []
        
        ##count = 0
        for line in f:
            try:
                if '#' in line:
                    G1 = nx.Graph()
                    #labels1 = {}
                    #node_colors1 = []
                    line = f.readline()
                    while 'v' in line:
                        data = line.split(' ')
                        node = np.int(data[1])
                        element = np.int(data[2])
                        #labels1[node] = CHEM_EL[element]['name']
                        #node_colors1.append(CHEM_EL[element]['color'])
                        #G1.add_node(node,label=labels1[node])
                        G1.add_node(node,label=CHEM_EL[element]['name'])
                        line = f.readline()
                    while 'e' in line:
                        data = line.split(' ')
                        vertice1 = np.int(data[1])
                        vertice2 = np.int(data[2])
                        G1.add_edge(vertice1,vertice2,weight=np.int(data[3]))
                        line = f.readline()
                    M_list_infreq.append(G1)
                    ##if isomorph_test(G1,G2):
                    ##    bin_matrix[count] = True
                    ##count += 1
            except ValueError:
                print("Invalid input:"+line)
        f.close()
        
        # top discriminative substructures:
        graphFile = "classifier/top_sg_"+str(i+1)+".gsp"
        f = open(graphFile,'r')
        SG_list = []
        
        for line in f:
            try:
                if '#' in line:
                    G2 = nx.Graph()
                    #labels2 = {}
                    #node_colors2 = []
                    line = f.readline()
                    while 'v' in line:
                        data = line.split(' ')
                        node = np.int(data[1])
                        element = np.int(data[2])
                        #labels2[node] = CHEM_EL[element]['name']
                        #node_colors2.append(CHEM_EL[element]['color'])
                        #G2.add_node(node,label=labels2[node])
                        G2.add_node(node,label=CHEM_EL[element]['name'])
                        line = f.readline()
                    while 'e' in line:
                        data = line.split(' ')
                        vertice1 = np.int(data[1])
                        vertice2 = np.int(data[2])
                        G2.add_edge(vertice1,vertice2,weight=np.int(data[3]))
                        line = f.readline()
                    SG_list.append(G2)
            except ValueError:
                print("Invalid input:"+line)
        f.close()

        ## coding training set:
        # freq molecules:
        for i_molecule in range(sampleSize):
            for i_substructure in range(len(SG_list)):
                training_data_X[i][i_molecule][i_substructure] = isomorph_test(M_list_freq[i_molecule],SG_list[i_substructure])
        
        # infreq molecules:
        for i_molecule in range(sampleSize):
            for i_substructure in range(len(SG_list)):
                training_data_X[i][sampleSize+i_molecule][i_substructure] = isomorph_test(M_list_infreq[i_molecule],SG_list[i_substructure])
    print("")
    
    if not os.path.exists('classifier'):
        os.mkdir('classifier')
    np.save('classifier/training_data_X.npy', training_data_X)
    np.save('classifier/training_data_Y.npy', training_data_Y)

def makeSamples(inputFile, outputStump, numSamplings, infreqSize, sampleSize):

    if not os.path.exists('sample_data'):
        os.mkdir('sample_data')
    for j in range(numSamplings):
        f = open(inputFile, 'r')
        outputFile = "sample_data/"+outputStump+"_sample"+str(j+1)+".gsp"
        f_sampleing = open(outputFile, 'w')
        
        sample_indexes = np.zeros(sampleSize,np.int)
        for i in range(sampleSize):
            number = np.random.randint(1,infreqSize)#avoids the 0-index problem
            while (number in sample_indexes):
                number = np.random.randint(1,infreqSize)
            sample_indexes[i] = number
            
        counter = 0
        for line in f:
            try:
                if ('t #' or 't#') in line:
                    data = line.split('#')
                    index = np.int(data[1])
                    if (index+1) in sample_indexes:#since gsp-indices begin at 0
                        counter += 1
                        f_sampleing.write(line)
                        while not (line == '\n'):
                            line = f.readline()
                            f_sampleing.write(line)
                        for i in range(sampleSize):
                            if sample_indexes[i] == (index+1):
                                sample_indexes[i] = -1
            except ValueError:
                print("Invalid input:"+line)
        f.close()
        f_sampleing.close()

def mineSubgraphs(inputStump, numSamplings, freqFile, numSubgraphs):
    print("Mining sample", end='', flush=True)
    if not os.path.exists('classifier'):
        os.mkdir('classifier')
    for j in range(numSamplings):
        nonFreqFile = "sample_data/"+inputStump+"_sample"+str(j+1)+".gsp"
        outputFile = "classifier/top_sg_"+str(j+1)+".gsp"
        f = open(outputFile, 'w')
        print(" "+str(j+1), end='', flush=True)
        subprocess.run(["./gspan", str(numSubgraphs), nonFreqFile, freqFile, "ig"], stdout=f)
        f.close()
    print("")

def hydrogenSuppression(inputFile, outputFile):
    f = open(inputFile, 'r')
    f_out = open(outputFile, 'w')
    count = 0

    for line in f:
        try:
            if '#' in line:
                count+=1
                f_out.write(line)
                line = f.readline()
                h_list = []
                while 'v' in line:
                    data = line.split(' ')
                    if np.int(data[2]) != 1:
                        f_out.write(line)
                    else:
                        h_list.append(data[1])
                    line = f.readline()
                while 'e' in line:
                    data = line.split(' ')
                    if (data[1] not in h_list) and (data[2] not in h_list):
                        f_out.write(line)                
                    line = f.readline()
                f_out.write('\n')
        except ValueError:
            print("Invalid input:"+line)
            
    f.close()
    f_out.close()

#total_features = 192

print("Opening "+sys.argv[1])

f = open(sys.argv[1], 'r')
freqRnd = random.randint(1000,100000)
freqSDF = "freq"+str(freqRnd)+".sdf"
f_freq = open(freqSDF, "w")
infreqRnd = random.randint(1000,100000)
infreqSDF = "infreq"+str(infreqRnd)+".sdf"
f_infreq = open(infreqSDF, "w")
#count = 0
#count_start = 0
#count_end = 0
#molecule_id = 0
assay = np.zeros(6,bool)
buff_str = ''
frequentHitters = 0
nonFrequentHitters = 0

print("Parsing sdf")

for line in f:
    try:
        buff_str += line
        #if '<PUBCHEM_CID>' in line:
        #    line = f.readline()
        #    molecule_id = np.int(line)
        #    buff_str += line
        if '<AID_623870>' in line:
            line = f.readline()
            #count += 1
            assay[0] = np.int(line)
            buff_str += line                   
        elif '<AID_624168>' in line:
            line = f.readline()
            #count += 1
            assay[1] = np.int(line)
            buff_str += line         
        elif '<AID_651704>' in line:
            line = f.readline()
            #count += 1
            assay[2] = np.int(line)
            buff_str += line           
        elif '<AID_651724>' in line:
            line = f.readline()
            #count += 1
            assay[3] = np.int(line)
            buff_str += line           
        elif '<AID_651725>' in line:
            line = f.readline()
            #count += 1
            assay[4] = np.int(line)
            buff_str += line           
        elif '<AID_743445>' in line:
            line = f.readline()
            #count += 1
            assay[5] = np.int(line)
            buff_str += line
        elif '<apol>' in line:
            line = f.readline()
            buff_str += line
            while not '$$$$' in line:
                line = f.readline()
                #count += 1
                buff_str += line
            #count_end = count                  
            if np.sum(assay) > 1: #molecule is frequent
                #print molecule_id, assay
                #print count_start, count_end, count_end-count_start
                #file_freq.write(str(count_start)+' '+str(count_end)+'\n')
                f_freq.write(buff_str)
                frequentHitters+=1
            else: #molecule is infrequent
                f_infreq.write(buff_str)
                #print molecule_id, assay
                #print count_start, count_end, count_end-count_start
                #file_infreq.write(str(count_start)+' '+str(count_end)+'\n')  
                ######count_start = count+1
                nonFrequentHitters+=1
            buff_str = ''
        #count += 1
        #buff_str += line
    except ValueError:
        print("Invalid input:"+line)
f.close()
f_freq.close()
f_infreq.close()

#print(str(frequentHitters)+":"+str(nonFrequentHitters)+" -> "+str(int(nonFrequentHitters/frequentHitters*1.16)+1))
print("Parsed "+str(frequentHitters)+" FHs and "+str(nonFrequentHitters)+" non-FHs -> non-FHs will be sampled "+str(int(nonFrequentHitters/frequentHitters*1.16)+1)+" times")

freqGSP = "freq"+str(freqRnd)+".gsp"
subprocess.run(["perl", "sdf2gsp_nolabels_mod.pl", freqSDF, freqGSP])

infreqGSP = "infreq"+str(infreqRnd)+".gsp"
subprocess.run(["perl", "sdf2gsp_nolabels_mod.pl", infreqSDF, infreqGSP])

freqHLess = "freq"+str(freqRnd)+"_hless.gsp"
infreqHLess = "infreq"+str(infreqRnd)+"_hless.gsp"

hydrogenSuppression(freqGSP, freqHLess)
hydrogenSuppression(infreqGSP, infreqHLess)

makeSamples(infreqHLess, "infreq"+str(infreqRnd)+"_hless", int(nonFrequentHitters/frequentHitters*1.16)+1, nonFrequentHitters, frequentHitters)

mineSubgraphs("infreq"+str(infreqRnd)+"_hless", int(nonFrequentHitters/frequentHitters*1.16)+1, freqHLess, int(sys.argv[2]))

encode("infreq"+str(infreqRnd)+"_hless", freqHLess, int(nonFrequentHitters/frequentHitters*1.16)+1, int(sys.argv[2]), frequentHitters)
#%%
